@@ -40,11 +40,16 @@
 #include <linux/workqueue.h>
 #include <linux/delay.h>
 #include <linux/sched.h>
+
 #include <linux/slab.h>
 
 #include <stdbool.h>
 #include <dahdi/kernel.h>
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
+#include <linux/sched/signal.h>
+#endif /* 4.11.0 */
+
 #include "wct4xxp/wct4xxp.h"	/* For certain definitions */
 
 #include "voicebus/voicebus.h"
